"""Path utilities relating to actions"""

import logging
import os
from typing import Optional, Set

logger = logging.getLogger(__name__)


def path_to_module(path: str) -> str:
    """
    Converts a file path into a module path.

    Parameters:
        path (str): The file path to convert. Example: '/a/b/c'

    Returns:
        str: The module path. Example: 'a.b.c'
    """
    # Strip leading and trailing slashes and split the path by slashes
    parts = path.strip("/").split("/")

    # Join the parts with dots to form the module path
    module_path = ".".join(parts)

    return module_path


def find_package_folder(
    rootdir: str, name: str, required_files: Optional[Set[str]] = None
) -> Optional[str]:
    """Find a package folder within a namespace."""
    if required_files is None:
        required_files = {"info.yaml"}
    try:
        # Split the provided name into namespace and package_folder
        namespace, package_folder = name.split("/")

        # Build the path for the namespace
        namespace_path = os.path.join(rootdir, namespace)

        # Check if the namespace directory exists
        if not os.path.isdir(namespace_path):
            return None

        # Traverse the namespace directory for the package_folder
        for root, dirs, _files in os.walk(namespace_path):
            if package_folder in dirs:
                package_path = os.path.join(root, package_folder)

                # Check for the presence of the required files
                folder_files = set(os.listdir(package_path))

                if required_files.issubset(folder_files):
                    # Return the full path of the package_folder if all checks out
                    return package_path

        # If package_folder is not found, return None
        return None

    except ValueError:
        logger.error("Invalid format. Please use 'namespace/package_folder'.")
        return None


def action_walker_path(module: str) -> str:
    """Accepts the module string of an action walker and returns a walker path string"""
    if not module:
        return ""

    module_parts = module.split(".")
    return f"/action/walker/{module_parts[-2]}/{module_parts[-1]}"


def action_webhook_path(module: str) -> str:
    """Accepts the module string of an action webhook walker and returns a walker path string"""
    if not module:
        return ""

    module_parts = module.split(".")
    return f"/action/webhook/{module_parts[-2]}/{module_parts[-1]}/{{agent_id}}/{{key}}"



"""File handling utils"""

from jvserve.lib.file_interface import (
    FILE_INTERFACE,
    file_interface,
    get_file_interface,
)

# ensure .jvdata is the root as it contains sensitive data which we don't
# want served by jvcli jvfileserve
jvdata_file_interface = (
    get_file_interface("") if FILE_INTERFACE == "local" else file_interface
)



"""node_get paginating node collections in Jivas."""

from jac_cloud.core.archetype import NodeAnchor


def node_get(query_filter: dict | None = None) -> list:
    """Retrieve a list of nodes from the 'node' collection based on the query filter."""

    if query_filter is None:
        return []

    # Execute the query
    cursor = NodeAnchor.Collection.find(query_filter)

    if cursor:
        nodes = [n.archetype for n in cursor]
        return nodes

    return []



"""Text formatting utilities"""

import logging
import re
import unicodedata
from typing import Any, Dict, List, Optional, Union

import ftfy

logger = logging.getLogger(__name__)


def list_to_phrase(lst: List[Any]) -> str:
    """Formats a list as a phrased list, e.g. ['one', 'two', 'three'] becomes 'one, two, and three'."""
    if not lst:
        return ""

    # Convert all elements to strings
    lst = list(map(str, lst))

    if len(lst) == 1:
        return lst[0]

    return ", ".join(lst[:-1]) + f", and {lst[-1]}"


def replace_placeholders(
    string_or_collection: Union[str, List[str]], placeholders: Dict[str, Any]
) -> Union[str, List[str], None]:
    """
    Replaces placeholders delimited by {{ and }} in a string or collection of strings with their corresponding values
    from a dictionary of key-value pairs. If a value in the dictionary is a list, it is formatted as a phrased list.
    Returns only items that have no remaining placeholders.
    """

    def replace_in_string(s: str, ph: Dict[str, Any]) -> Optional[str]:
        # Replace placeholders in a single string
        for key, value in ph.items():
            if isinstance(value, list):
                value = list_to_phrase(value)
            s = s.replace(f"{{{{{key}}}}}", str(value))
        return s if not re.search(r"{{.*?}}", s) else None

    if isinstance(string_or_collection, str):
        return replace_in_string(string_or_collection, placeholders)
    elif isinstance(string_or_collection, list):
        # Process each string in the list and return those with no placeholders left
        return [
            result
            for string in string_or_collection
            if (result := replace_in_string(string, placeholders)) is not None
        ]
    else:
        raise TypeError("Input must be a string or a list of strings.")


def escape_string(input_string: str) -> str:
    """
    Escapes curly braces in the input string by converting them to double curly braces.

    Args:
        input_string: The string to be escaped.

    Returns:
        The escaped string with all curly braces doubled.
    """
    if not isinstance(input_string, str):
        logger.error(f"Error expect string: {input_string}")
        return ""
    return input_string.replace("{", "{{").replace("}", "}}")


def normalize_text(text: str) -> str:
    """
    Normalizes a string by:
    - Stripping whitespace
    - Fixing text encoding
    - Removing non-word characters

    Args:
        text: Input string
    Returns:
        Normalized string
    """
    text = text.strip().replace("\\n", "\n")
    text = ftfy.fix_text(text)
    return re.sub(r"\W+", "", text)


def to_snake_case(title: str, ascii_only: bool = True) -> str:
    """
    Converts a string to a safe, lowercase snake_case suitable for machine names.

    Args:
        title: The input string/title
        ascii_only: If True, strips out non-ascii characters

    Returns:
        A clean, lowercase, snake_case string
    """
    s = title.strip()
    if ascii_only:
        # Normalize unicode
        s = unicodedata.normalize("NFKD", s)
        s = s.encode("ascii", "ignore").decode("ascii")
    else:
        # Remove combining marks but keep unicode letters
        s = unicodedata.normalize("NFKC", s)
    # Replace non-alphanumeric with underscores
    s = re.sub(r"[^a-zA-Z0-9]+", "_", s)
    # Lowercase it
    s = s.lower()
    # Remove leading/trailing and consecutive underscores
    s = re.sub(r"_{2,}", "_", s)
    s = s.strip("_")
    return s


def clean_text(text: Optional[str], force_ascii: bool = False) -> str:
    """
    Sanitizes text to prevent encoding issues (e.g., UnicodeEncodeError).

    Args:
        text: Input string (or None)
        force_ascii: If True, replaces non-ASCII chars with '?'

    Returns:
        Cleaned string, or empty string if input is None
    """
    if text is None:
        return ""

    # Remove Unicode line/paragraph separators
    text = re.sub(r"[\u2028\u2029]", " ", text)

    if force_ascii:
        # Encode to ASCII, replacing non-ASCII chars with '?'
        text = text.encode("ascii", errors="replace").decode("ascii")

    return text.strip()



"""MIME types package"""

import logging
import mimetypes
import os
from typing import Dict, Optional

import requests

logger = logging.getLogger(__name__)


def get_mime_type(
    file_path: Optional[str] = None,
    url: Optional[str] = None,
    mime_type: Optional[str] = None,
) -> Optional[Dict[str, str]]:
    """Determines the MIME type and categorizes it into common file types."""
    detected_mime_type = None

    if file_path:
        detected_mime_type, _ = mimetypes.guess_type(file_path)
    elif url:
        try:
            response = requests.head(url, allow_redirects=True)
            detected_mime_type = response.headers.get("Content-Type")
        except requests.RequestException as e:
            logger.error(f"Error making HEAD request: {e}")
            return None
    else:
        detected_mime_type = mime_type

    # MIME type categories
    image_mime_types = [
        "image/jpeg",
        "image/png",
        "image/gif",
        "image/bmp",
        "image/webp",
        "image/tiff",
        "image/svg+xml",
        "image/x-icon",
        "image/heic",
        "image/heif",
        "image/x-raw",
    ]
    document_mime_types = [
        "application/pdf",
        "application/msword",
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.ms-excel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        "application/vnd.ms-powerpoint",
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "text/plain",
        "text/csv",
        "text/html",
        "application/rtf",
    ]
    audio_mime_types = [
        "audio/mpeg",
        "audio/wav",
        "audio/ogg",
        "audio/flac",
        "audio/aac",
        "audio/mp3",
        "audio/webm",
        "audio/amr",
        "audio/midi",
    ]
    video_mime_types = [
        "video/mp4",
        "video/mpeg",
        "video/ogg",
        "video/webm",
        "video/quicktime",
        "video/x-msvideo",
        "video/x-matroska",
    ]

    if detected_mime_type is None or detected_mime_type == "binary/octet-stream":
        if file_path:
            _, file_extension = os.path.splitext(file_path)
        elif url:
            _, file_extension = os.path.splitext(url)
        else:
            file_extension = ""
        detected_mime_type = mimetypes.types_map.get(file_extension.lower())

    # Categorize MIME type
    if detected_mime_type in image_mime_types:
        return {"file_type": "image", "mime": detected_mime_type}
    elif detected_mime_type in document_mime_types:
        return {"file_type": "document", "mime": detected_mime_type}
    elif detected_mime_type in audio_mime_types:
        return {"file_type": "audio", "mime": detected_mime_type}
    elif detected_mime_type in video_mime_types:
        return {"file_type": "video", "mime": detected_mime_type}
    else:
        logger.error(f"Unsupported MIME Type: {detected_mime_type}")
        return None
